#!/usr/bin/env python3
# pylint:

import os
import sys
import re
import configparser
import argparse
import json
import urllib.parse
import requests
import jinja2
from enc_hosts.version import version

# ---------------------------------------------------------------------------
#                                                               FUNCTIONS
# ---------------------------------------------------------------------------

# pylint: disable=unused-argument
def print_response(r, *args, **kwargs):
  print("Receiving from: {}".format(r.url), file=sys.stderr)

def query(url, query, auth, tls_verify=True):

  qurl = url + query
  try:
    r = requests.get(qurl, verify=tls_verify, auth=auth, hooks={'response':print_response})
  except Exception as e:
    print(e.traceback.format_exc())
  return r.json()

def cache_valid(path):
  if os.path.exists(path):
    return True
  return False

def read_from_cache(path):
  with open(path, 'r') as file:
    return json.loads(file.read())

def write_to_cache(path, hosts):
  with open(path, 'w') as file:
    file.write(json.dumps(hosts))

def query_facts_from_hostid(auth, base_url, host_fqdn, tls_verify=True):
  page = 1
  facts = {}
  r = query(base_url,'{}?page={}'.format(
      'hosts/'+ str(host_fqdn) + '/facts', page), auth, tls_verify)
  while r['results']:
    facts.update(r['results'][host_fqdn])
    page += 1
    r = query(base_url,'{}?page={}'.format(
        'hosts/'+ str(host_fqdn) + '/facts', page), auth, tls_verify)
  return facts

def query_from_enc(auth, base_url, querystr, return_facts, tls_verify=True):

  # get list of hosts
  done = False
  page = 1
  hosts = {}
  while not done:
    r = query(base_url, '{}&page={}'.format(querystr, page), auth, tls_verify)

    # add results to hosts dictionary
    ## if return_facts is true, then do a second query for facts
    for host in r['results']:
      hosts[host['name']] = host
      host['x_params_map'] = {}
      for parameter in host['all_parameters']:
        host['x_params_map'][parameter['name']] = parameter['value']
      # if they've asked for the facts, run a second API call for this host
      if return_facts:
        facts = query_facts_from_hostid(auth, base_url, host['name'], tls_verify)
        host['facts'] = facts
    # where are we with the pagination
    if r['per_page'] * r['page'] >= r['subtotal']:
      done = True
    else:
      page += 1
  return hosts

# ---------------------------------------------------------------------------
#                                                               JINJA FILTERS
# ---------------------------------------------------------------------------

def regex(str, regex):
  regexc = re.compile(regex)
  return regexc.match(str)

# ---------------------------------------------------------------------------
#                                                               MAIN
# ---------------------------------------------------------------------------

def main():

  # program name
  _tag = "enc-hosts"

  # program description
  description = '''
  Pull host information from ENC (enterprise node classifier) and run through
  given template.
  '''
  epilog = "Version {}.  For more information please visit project homepage: " \
    "https://git.computecanada.ca/dleske/enc-hosts".format(version)

  # get command-line options
  # Usage: $0 [-c config]
  parser = argparse.ArgumentParser(prog=_tag, description=description, epilog=epilog)
  parser.add_argument("-c", "--config", type=str,
                      help="Configuration file",
                      default=__file__ + ".conf")
  parser.add_argument("--use-cache",
                      help="Use cached results if available",
                      action="store_true")
  parser.add_argument("--search", help="Host search string")
  parser.add_argument("--suffix", type=str,
                      help="Suffix to apply to rendered templates",
                      default=".out")
  parser.add_argument("--facts",
                      help="Also pull facts for each host (CAUTION: very resource intensive)",
                      action="store_true")
  parser.add_argument("templates", metavar='TEMPLATE',
                      help="Template(s) to render", nargs="+")
  args = parser.parse_args()

  # initialize configuration
  config = configparser.ConfigParser(delimiters='=')
  config.read(args.config)

  # set up to skip TLS verification if so configured
  tls_verify = True
  if config['main']['skip_verification']:
    # pylint: disable=import-outside-toplevel
    import urllib3
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
    tls_verify = False

  # build query
  base_url = config['main']['server_url'] + '/api/'
  auth = (config['main']['username'], config['main']['password'])
  query = "hosts?include=all_parameters"
  if args.search:
    encoded_search = urllib.parse.quote_plus(args.search)
    query += '&search=' + encoded_search
    cachefile = encoded_search
  else:
    cachefile = 'hosts'

  # determine cachepath
  cachepath = None
  if 'cache_dir' in config['main']:
    cachedir = os.path.expanduser(config['main']['cache_dir'])
    if os.path.exists(cachedir):
      cachepath = '{}/{}'.format(cachedir, cachefile)

  # if using cache and cache is valid
  if args.use_cache and cache_valid(cachepath):
    print("Reading from cache", file=sys.stderr)
    hosts = read_from_cache(cachepath)
  else:
    hosts = query_from_enc(auth, base_url, query, args.facts, tls_verify)

    # if caching
    if cachepath:
      write_to_cache(cachepath, hosts)

  # template loading
  templatedir = config['main'].get('template_dir', 'templates')
  templateloader = jinja2.FileSystemLoader(templatedir)

  # templating environment
  jenv = jinja2.Environment(loader=templateloader)

  # add custom tests
  jenv.tests['regex'] = regex

  # loop through given templates
  for template in args.templates:
    template_fn = template + '.jinja'
    rendered_fn = template + args.suffix

    print("Rendering template {} to {}".format(template_fn, rendered_fn), file=sys.stderr)

    jtmpl = jenv.get_template(template_fn)
    with open(rendered_fn, 'w') as rendered:
      rendered.write(jtmpl.render(hosts=hosts))

# invoke main function
if __name__ == '__main__':
  sys.exit(main())
